#include "kronprd.h"
using namespace scl;
using namespace std;

int main()
{
  cout << starbox("/A, B, K(A,B)//");
  realmat A(2,3);
  for (INTEGER i=1; i<=A.size(); ++i) A[i] = i;
  realmat B(3,2,1.0);
  kronprd K(A,B);
  cout << A << B << K << '\n';

  cout << starbox("/Test trealmat constructors//") << '\n' ;
  A = T(A);  B = T(B);
  cout << K - kronprd(T(A),T(B));
  A = T(A);
  cout << K - kronprd(A,T(B));
  A = T(A);  B = T(B);
  cout << K - kronprd(T(A),B);
  
  cout << starbox("/Test indexing by strings and intvecs//") << '\n' ;
  cout << "K(1,\"\") = " << K(1,"") << '\n';
  cout << "K(3,\"\") = " << K(3,"") << '\n';
  cout << "K(\"\",6) = " << K("",6) << '\n';
  cout << "K(\"\",2) = " << K("",2) << '\n';
  cout << "K(\"1,3\",\"\") = " << K("1,3","") << '\n';
  cout << "K(\"\",\"6,2\") = " << K("","6,2") << '\n';
  cout << "K(\"2:4\",\"6,2\") = " << K("2:4","6,2") << '\n';
  cout << "K(seq(2,4),seq(6,2)) = " << K(seq(2,4),seq(6,2)) << '\n';
  
  cout << starbox("/Convert a null matrix//");
  kronprd K0;
  realmat R0;
  R0 = K0;
  cout << R0 << '\n';

  cout << starbox("/Test copy constructor//");
  kronprd K1 = K;
  cout << K1 << '\n';

  cout << starbox("/Test all the gets//") << '\n';
  cout << "     rowsA = " << fmt('d', 5, K.get_rowsA() ) <<'\n';
  cout << "     colsA = " << fmt('d', 5, K.get_colsA() ) <<'\n';
  cout << K.get_A() << '\n';
  cout << "     rowsB = " << fmt('d', 5, K.get_rowsB() ) <<'\n';
  cout << "     colsB = " << fmt('d', 5, K.get_colsB() ) <<'\n';
  cout << K.get_B() << '\n';
  cout << "     rows = " << fmt('d', 5, K.get_rows() ) <<'\n';
  cout << "     cols = " << fmt('d', 5, K.get_cols() ) <<'\n';
  cout << "     size = " << fmt('d', 5, K.size() ) <<'\n';
  cout << "     len = " << fmt('d', 5, K.get_len() ) <<'\n';
  cout << K << '\n';
  
  cout << starbox("/Test operator=//") << '\n';
  K0 = K;
  cout << K0 << '\n';
  
  cout << starbox("/Test indexing//") << '\n';
  for (INTEGER i=1; i<=K.size(); ++i) {
    cout << fmt('d',10,i) << fmt('f',10,5,K[i]) << '\n';
  }
  for (INTEGER j=1; j<=K.get_rows(); ++j) {
    for (INTEGER i=1; i<=K.get_cols(); ++i) {
      cout << fmt('d',10,i) << fmt('d',5,j) << fmt('f',10,5,K(i,j)) << '\n';
    }
  }

  cout << starbox("/Test indexing with check//") << '\n';
  for (INTEGER i=1; i<=K.size(); ++i) {
    cout << fmt('d',10,i) << fmt('f',10,5,K.check(i)) << '\n';
  }
  for (INTEGER j=1; j<=K.get_rows(); ++j) {
    for (INTEGER i=1; i<=K.get_cols(); ++i) {
      cout << fmt('d',10,i) << fmt('d',5,j) << fmt('f',10,5,K.check(i,j)) 
           << '\n';
    }
  }

  cout << starbox("/Test operator*//") << '\n';
  realmat A3(2,3);
  realmat B3(1,2);
  realmat A4(3,3);
  realmat B4(2,1);
  for (INTEGER i=1; i<=A3.size(); ++i) A3[i]=i;
  for (INTEGER i=1; i<=B3.size(); ++i) B3[i]=i;
  for (INTEGER i=1; i<=A4.size(); ++i) A4[i]=i;
  for (INTEGER i=1; i<=B4.size(); ++i) B4[i]=i;
  kronprd K3(A3,B3);
  kronprd K4(A4,B4);
  cout << A3 << B3 << K3 << A4 << B4 << K4 << '\n';
  kronprd K5 = K3*K4;
  cout << K5.get_A() << K5.get_B() << '\n';
  realmat R3 = K3;
  cout << K5 << '\n';
  cout << R3*K4 << '\n';
  realmat R4 = K4;
  cout << K3*R4 << '\n';

  cout << starbox("/Test operator+//") << '\n';
  cout << K3 << +K3 << '\n';
  cout << K3 << K3+K3 << K3+K3+K3 << '\n';
  cout << R3 << R3+K3 << R3+K3+K3 << '\n';
  cout << R3 << K3+R3 << K3+(K3+R3) << '\n';

  cout << starbox("/Test operator-//") << '\n';
  cout << K3 << -K3 << '\n';
  cout << K3 << K3-K3 << K3-K3-K3 << '\n';
  cout << R3 << R3-K3 << R3-K3-K3 << '\n';
  cout << R3 << K3-R3 << K3-(K3-R3) << '\n';

  cout << starbox("/Test invpsd//");
  kronprd K6(A3*T(A3),B3*T(B3));
  INTEGER rank;
  cout << K6 << invpsd(K6,rank)<< '\n'; 
  cout << "      rank = " << fmt('d',5,rank) << '\n';
  cout << K6 << inv(K6,rank) << '\n'; 
  cout << "      rank = " << fmt('d',5,rank) << '\n';
  cout << K6 << ginv(K6,rank) << '\n'; 
  cout << "      rank = " << fmt('d',5,rank) << '\n';
  
  cout << starbox("/Test unary operator-//") << '\n';
  realmat A7(4,2,1.0);
  realmat B7(2,3,2.0);;
  A7(1,2) = A7(3,1) = A7(2,2) = 0.0;
  B7(1,2) = B7(1,3) = B7(2,3) = 0.0;
  kronprd K7(A7,B7);
  cout << A7 << B7;
  cout << K7 << -K7 << '\n';
  kronprd K8 = -K7;
  cout << K8.get_A() << K8.get_B();

  return 0;
}
